#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Table for cisco_apic network extra contracts extension attributes

Revision ID: f28141ea1bbf
Revises: b76dc22f2e23
Create Date: 2019-12-20 13:08:03.603312

"""

# revision identifiers, used by Alembic.
revision = 'f28141ea1bbf'
down_revision = 'b76dc22f2e23'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'apic_aim_network_extra_contracts',
        sa.Column('network_id', sa.String(36), nullable=False),
        sa.Column('contract_name', sa.String(64), nullable=False),
        sa.Column('provides', sa.Boolean, nullable=False),
        sa.ForeignKeyConstraint(
            ['network_id'], ['networks.id'],
            name='apic_aim_network_contract_extn_fk_network',
            ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('network_id', 'contract_name', 'provides')
    )


def downgrade():
    pass
